/*
**==============================================================================
**
** Open Management Infrastructure (OMI)
**
** Copyright (c) Microsoft Corporation
** 
** Licensed under the Apache License, Version 2.0 (the "License"); you may not 
** use this file except in compliance with the License. You may obtain a copy 
** of the License at 
**
**     http://www.apache.org/licenses/LICENSE-2.0 
**
** THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
** KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED 
** WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABLITY OR NON-INFRINGEMENT. 
**
** See the Apache 2 License for the specific language governing permissions 
** and limitations under the License.
**
**==============================================================================
*/

#ifndef _pal_memory_h
#define _pal_memory_h

#include "nits/base/nits.h"

PAL_BEGIN_EXTERNC

/* Defining FASTER macro disables malloc injection
   to reduce overhead for manual performance benchmarking.
   It is not defined durring normal builds.
*/
#ifndef FASTER

PAL_INLINE void *malloc_Injected(
    size_t length,
    NitsCallSite cs)
{
    if (NitsShouldFault(cs, NitsAutomatic))
        return NULL;

    return malloc(length);
}

#ifdef malloc
# undef malloc
#endif
#define malloc(n) malloc_Injected(n, NitsHere())

PAL_INLINE void *calloc_Injected(
    size_t num,
    size_t size,
    NitsCallSite cs)
{
    if (NitsShouldFault(cs, NitsAutomatic))
        return NULL;

    return calloc(num, size);
}

#ifdef calloc
# undef calloc
#endif
#define calloc(n, s) calloc_Injected(n, s, NitsHere())

PAL_INLINE void *realloc_Injected(
    _In_ void *originalBuffer,
    size_t length,
    NitsCallSite cs)
{
    if (NitsShouldFault(cs, NitsAutomatic))
        return NULL;

    return realloc(originalBuffer, length);
}

#ifdef realloc
# undef realloc
#endif
#define realloc(p, n) realloc_Injected(p, n, NitsHere())

#endif

/* 
   // Required for PAL_Malloca
   //
   // (FUTURE) Defines a stack buffer for PAL_Malloca. On platforms that don't
   // support 'alloca', reserves some space for PAL_Malloca to allocate
   // from before falling back to malloc.
   PAL_MALLOCA_STACK(initialSize) 
 
   // attempts to allocate from the stack via 'alloca' if available,
   // otherwise from malloc. Must be freed by PAL_Freea.
   void* PAL_Malloca(size_t sz)
 
   // free an Alloca pointer. p may be null.
   void PAL_Freea(void *p)
*/
#if defined(_MSC_VER) && !defined(PAL_NO_ALLOCA)
#include <malloc.h>
#define PAL_MALLOCA_STACK(initialSize) char _Pal_Malloca_Buffer[(initialSize)>0]
#define PAL_Malloca(sz) (&_Pal_Malloca_Buffer, NitsShouldFault(NitsHere(), NitsAutomatic) ? NULL : _malloca((sz)))
#define PAL_Freea(p) _freea((p))
#elif defined(__GNUC__) && !defined(PAL_NO_ALLOCA)
#include <alloca.h>
#define PAL_MALLOCA_STACK(initialSize) char _Pal_Malloca_Buffer[(initialSize)>0]
#define PAL_Malloca(sz) (&_Pal_Malloca_Buffer, NitsShouldFault(NitsHere(), NitsAutomatic) ? NULL : alloca((sz)))
#define PAL_Freea(p)
#else 
/* TODO/FIXME: add bookkeeping to support allocating from a fixed sized stack
   buffer, failing over to malloc when it runs out. */
#define PAL_MALLOCA_STACK(initialSize) char _Pal_Malloca_Buffer[(initialSize)>0]
#define PAL_Malloca(sz) (&_Pal_Malloca_Stack, malloc((sz)))
#define PAL_Freea(p) free((p))
#endif

PAL_END_EXTERNC

#endif /* _pal_memory_h */
