/*
**==============================================================================
**
** Open Management Infrastructure (OMI)
**
** Copyright (c) Microsoft Corporation
** 
** Licensed under the Apache License, Version 2.0 (the "License"); you may not 
** use this file except in compliance with the License. You may obtain a copy 
** of the License at 
**
**     http://www.apache.org/licenses/LICENSE-2.0 
**
** THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
** KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED 
** WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABLITY OR NON-INFRINGEMENT. 
**
** See the Apache 2 License for the specific language governing permissions 
** and limitations under the License.
**
**==============================================================================
*/

#include "filter.h"
#include <base/log.h>
#include <pal/strings.h>
#include <pal/format.h>
#include <indication/common/indicommon.h>

static const MI_Uint32 _MAGIC = 0xDEADBEEF;

/*
 * Structure and handler for WQL queries.
 */
typedef struct _IndicationFilter_WQL
{
    IndicationFilter filterBase;

    WQL* wql;
} IndicationFilter_WQL;

/*
 * Structure and handler for CQL queries.  Right now, CQL queries are handled
 * exactly like WQL queries.
 */
typedef struct _IndicationFilter_CQL
{
    IndicationFilter filterBase;

    WQL* wql;
} IndicationFilter_CQL;

/*
 * Additional derived structures for new filter types go here.
 */

/*
 * Internal IndicationFilterFT function definitions
 */
MI_Result _FilterFunc_WQL( 
    IndicationFilter* self, 
    const MI_Instance* indication,
    MI_Boolean* isMatch )
{
    int r = 0;

    r = WQL_Eval(((IndicationFilter_WQL*)self)->wql, WQL_LookupInstanceProperty, (void*)indication);
    if (r == 0)
    {
        /* Instance matched the query */
        *isMatch = MI_TRUE;
    }
    else
    {
        /* Mismatch */
        *isMatch = MI_FALSE;
    }
    return MI_RESULT_OK;
}

MI_Result _FilterFunc_CQL( 
    IndicationFilter* self, 
    const MI_Instance* indication,
    MI_Boolean* isMatch)
{
    int r = 0;

    r = WQL_Eval(((IndicationFilter_CQL*)self)->wql, WQL_LookupInstanceProperty, (void*)indication);
    if (r == 0)
    {
        /* Instance matched the query */
        *isMatch = MI_TRUE;
    }
    else
    {
        /* Mismatch */
        *isMatch = MI_FALSE;
    }
    return MI_RESULT_OK;
}

void _Destroy_WQL(
    _In_ IndicationFilter* self )
{
    WQL_Delete( ((IndicationFilter_WQL*)self)->wql );
}

void _Destroy_CQL(
    _In_ IndicationFilter* self )
{
    WQL_Delete( ((IndicationFilter_CQL*)self)->wql );
}

IndicationFilterFT _filterFT_WQL =
{
    _FilterFunc_WQL,
    _Destroy_WQL
};

IndicationFilterFT _filterFT_CQL =
{
    _FilterFunc_CQL,
    _Destroy_CQL
};

static MI_Result MI_CALL _Evaluate(
    _In_ const MI_Filter* self, 
    _In_ const MI_Instance* instance,
    _Out_ MI_Boolean* result)
{
    IndicationFilter* filter = (IndicationFilter*)self;

    if (!filter || filter->magic != _MAGIC || !instance || !result)
    {
        trace_IndicationFilter_EvaluateWithNullInput();
        return MI_RESULT_INVALID_PARAMETER;
    }

    return filter->ft->Filter( filter, instance, result );
}

static MI_Result MI_CALL _GetExpression(
    _In_ const MI_Filter* self, 
    _Outptr_result_maybenull_z_ const MI_Char** queryLang, 
    _Outptr_result_maybenull_z_ const MI_Char** queryExpr)
{
    IndicationFilter* filter = (IndicationFilter*)self;

    if (!filter || filter->magic != _MAGIC)
    {
        trace_IndicationFilter_GetExpressionWithNullInput();
        return MI_RESULT_INVALID_PARAMETER;
    }

    if (queryLang)
    {
        *queryLang = filter->msg->language;
    }

    if (queryExpr)
    {
        *queryExpr = filter->msg->filter;
    }
    return MI_RESULT_OK;
}

MI_FilterFT __mi_FilterFT =
{
    _Evaluate,
    _GetExpression
};

_Use_decl_annotations_
IndicationFilter* IndicationFilter_New( 
    SubscribeReq* msg )
{
    if (NULL == msg->language) /* protects against calls with NULL values to the comparison functions */
    {
        trace_UnsupportedQueryLanguage(tcs(msg->language));
        return NULL;
    }
    else if (Tcscasecmp(msg->language, QUERY_LANGUAGE_WQL) == 0)
    {
        IndicationFilter_WQL* filter = (IndicationFilter_WQL*)Batch_GetClear(msg->base.base.batch, sizeof(IndicationFilter_WQL));

        if(!filter)
        {
            trace_OutOfMemory();
            return NULL;
        }
        /* Initialize "Base" struct */
        filter->filterBase.base.ft = &__mi_FilterFT;
        filter->filterBase.magic = _MAGIC;
        filter->filterBase.msg = msg;
        filter->filterBase.ft = &_filterFT_WQL;

        /* Initialize WQL extension */
        filter->wql = WQL_Parse(msg->filter, msg->base.base.batch, WQL_DIALECT_WQL);
        if (!filter->wql)
        {
            trace_InvalidQueryExpression(tcs(msg->filter));
            /* filter will be freed when the Batch destructs */
            return NULL;
        }
        
        return (IndicationFilter*)filter;
    }
    else if (Tcscasecmp(msg->language, QUERY_LANGUAGE_CQL) == 0)
    {
        IndicationFilter_CQL* filter = (IndicationFilter_CQL*)Batch_GetClear(msg->base.base.batch, sizeof(IndicationFilter_CQL));

        if(!filter)
        {
            trace_OutOfMemory();
            return NULL;
        }

        /* Initialize "Base" struct */
        filter->filterBase.base.ft = &__mi_FilterFT;
        filter->filterBase.magic = _MAGIC;
        filter->filterBase.msg = msg;
        filter->filterBase.ft = &_filterFT_CQL;

        /* Initialize WQL extension since CQL covers the same type of supported
         * expressions. */
        filter->wql = WQL_Parse(msg->filter, msg->base.base.batch, WQL_DIALECT_CQL);
        if (!filter->wql)
        {
            trace_InvalidQueryExpression(tcs(msg->filter));
            /* filter will be freed when the Batch destructs */
            return NULL;
        }
        
        return (IndicationFilter*)filter;
    }    
    else
    {
        trace_UnsupportedQueryLanguage(tcs(msg->language));
        return NULL;
    }
}

_Use_decl_annotations_
void IndicationFilter_Destroy(
    IndicationFilter* self )
{
    DEBUG_ASSERT( self );

    /* Clean up specific to derived structs */
    self->ft->Destroy( self );

    /* Common clean up */
    self->magic = 0xFFFFFFFF;
}

_Use_decl_annotations_
MI_Result IndicationFilter_Filter(
    IndicationFilter* self, 
    const MI_Instance* indication,
    MI_Boolean* isMatch )
{
    return self->ft->Filter( self, indication, isMatch );
}

