/*
**==============================================================================
**
** Open Management Infrastructure (OMI)
**
** Copyright (c) Microsoft Corporation
** 
** Licensed under the Apache License, Version 2.0 (the "License"); you may not 
** use this file except in compliance with the License. You may obtain a copy 
** of the License at 
**
**     http://www.apache.org/licenses/LICENSE-2.0 
**
** THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
** KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED 
** WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABLITY OR NON-INFRINGEMENT. 
**
** See the Apache 2 License for the specific language governing permissions 
** and limitations under the License.
**
**==============================================================================
*/

#include <ut/ut.h>
#include <pal/palcommon.h>
#include <pal/strings.h>
#include <provmgr/filter.h>
#include <indication/common/indicommon.h>
#include <provmgr/indicationSchema.h>
#include <provmgr/CIM_InstCreation.h>
#include <provmgr/CIM_InstDeletion.h>

SubscribeReq* subscribeReq = NULL;

MI_ConstString defaultIndicationFilter = MI_T("SELECT * FROM CIM_InstCreation WHERE IndicationIdentifier = \"MSFT:CreationIndicationIdentifier\"");
MI_ConstString indicationLanguageWql = MI_T("WQL");
MI_ConstString indicationLanguageCql = MI_T("CQL");

/*
**==============================================================================
**
** Setup function, it is called before running each test case
** NOTE: for "nits -fault" run, this function will be called once and only once
** for each test case.
**
**==============================================================================
*/
NitsSetup(Test_IndicationFilter_SetupSubscribeReq)
{
    subscribeReq = SubscribeReq_New( 0, 0 );

    NitsAssertOrReturn(NULL != subscribeReq, PAL_T("Failed to initialize SubscribeReq"));

    subscribeReq->targetType = SUBSCRIP_TARGET_DEFAULT;
    subscribeReq->filter = defaultIndicationFilter;
    subscribeReq->language = indicationLanguageWql;
    subscribeReq->subscriptionID = 1;
}
NitsEndSetup

/*
**==============================================================================
**
** Cleanup function, it is called after run each test case
** NOTE: for "nits -fault" run, this function will be called once and only once
** for each test case.
**
**==============================================================================
*/
NitsCleanup(Test_IndicationFilter_SetupSubscribeReq)
{
    if (subscribeReq)
    {
        SubscribeReq_Release( subscribeReq );
        subscribeReq = NULL;
    }
}
NitsEndCleanup

NitsSetup1(Test_IndicationFilter_SetupSubscribeReq_as_WQL, NitsEmptyStruct,
           Test_IndicationFilter_SetupSubscribeReq, NitsEmptyValue)
{
    subscribeReq->language = indicationLanguageWql;
}
NitsEndSetup

NitsSetup1(Test_IndicationFilter_SetupSubscribeReq_as_CQL, NitsEmptyStruct,
           Test_IndicationFilter_SetupSubscribeReq, NitsEmptyValue)
{
    subscribeReq->language = indicationLanguageCql;
}
NitsEndSetup

NitsSplit2(Test_IndicationFilter_SetupSubscribeReq_BothWays,NitsEmptyStruct,
           Test_IndicationFilter_SetupSubscribeReq_as_WQL,
           Test_IndicationFilter_SetupSubscribeReq_as_CQL)
NitsEndSplit

// IndicationFilter_New with invalid query SubscribeReq->language
// IndicationFilter_Destroy resets magic
NitsTest1(Test_IndicationFilter_New_with_invalid_query_language, Test_IndicationFilter_SetupSubscribeReq, NitsEmptyValue)
{

    IndicationFilter* filter = NULL;
    subscribeReq->language = MI_T("InvalidQueryLanguage");

    filter = IndicationFilter_New( subscribeReq );

    NitsAssert(NULL == filter, PAL_T("Failed to initialize IndicationFilter with InvalidQueryLanguage"));
}
NitsEndTest

// IndicationFilter_New with WQL || CQL with invalid SubscribeReq->filter
// IndicationFilter_Destroy resets magic
NitsTest1(Test_IndicationFilter_New_with_invalid_query_filter, Test_IndicationFilter_SetupSubscribeReq_BothWays, NitsEmptyValue)
{
    subscribeReq->filter = MI_T("* FROM TEST_DummyIndicationClass");  // NO "SELECT"

    IndicationFilter* filter = IndicationFilter_New( subscribeReq );

    NitsAssert(NULL == filter, PAL_T("Failed to initialize IndicationFilter with InvalidQueryLanguage"));
    NitsIgnoringError(); // negative test case and same error in OOM or otherwise
}
NitsEndTest

// IndicationFilter_New with WQL || CQL initialized
// IndicationFilter_Destroy resets magic
NitsTest1(Test_IndicationFilter_New, Test_IndicationFilter_SetupSubscribeReq_BothWays, NitsEmptyValue)
{
    IndicationFilter* filter = IndicationFilter_New( subscribeReq );

    NitsAssertOrReturn(NULL != filter, PAL_T("Failed to initialize IndicationFilter with InvalidQueryLanguage"));
    NitsAssert(0xDEADBEEF == filter->magic, PAL_T("IndicationFilter: incorrect magic #"));
    NitsAssert(NULL != filter->ft, PAL_T("IndicationFilter: function table NOT set"));
    NitsAssert(subscribeReq == filter->msg, PAL_T("IndicationFilter: incorrect message"));

    IndicationFilter_Destroy( filter );

    NitsAssert(0xFFFFFFFF == filter->magic, PAL_T("IndicationFilter: magic # not reset following destroy"));
}
NitsEndTest

// IndicationFilter_Filter match with WQL || CQL
// MI_Filter Evaluate with WQL || CQL
NitsTest1(Test_IndicationFilter_Filter_match, Test_IndicationFilter_SetupSubscribeReq_BothWays, NitsEmptyValue)
{
    MI_Boolean isMatch = MI_FALSE;
    CIM_InstCreation indication;

    IndicationFilter* filter = IndicationFilter_New( subscribeReq );

    NitsAssertOrReturn(NULL != filter, PAL_T("Failed to initialize IndicationFilter with InvalidQueryLanguage"));

    if ( NitsAssert( MI_RESULT_OK == Instance_Construct( &indication.__instance, &CIM_InstCreation_rtti, NULL ), PAL_T("Instance_Construct failed") ) )
    {
        MI_Value value;
        value.string = (MI_Char*)PAL_T("MSFT:CreationIndicationIdentifier");
        if ( ! NitsAssert(MI_RESULT_OK == MI_Instance_SetElement( &indication.__instance, PAL_T("IndicationIdentifier"), &value, MI_STRING, 0 ), PAL_T("Set key failed on indication")))
        {
            IndicationFilter_Destroy( filter );
            NitsReturn;
        }

        // Test the instance against the internal version of the call
        NitsAssert( MI_RESULT_OK == IndicationFilter_Filter( filter, &indication.__instance, &isMatch ), PAL_T("Filter failed") );
        NitsAssert( MI_TRUE == isMatch, PAL_T("Filter match failed") );

        // Also test same instance via external MI call
        NitsAssert( MI_RESULT_OK == MI_Filter_Evaluate( &filter->base, &indication.__instance, &isMatch ), PAL_T("Filter failed")  );
        NitsAssert( MI_TRUE == isMatch, PAL_T("Filter match failed") );
        NitsAssert( MI_RESULT_OK == MI_Instance_Destruct( &indication.__instance ), PAL_T("indication destruction failed") );
    }
       

    IndicationFilter_Destroy( filter );
    NitsAssert(0xFFFFFFFF == filter->magic, PAL_T("IndicationFilter: magic # not reset following destroy"));

}
NitsEndTest

// IndicationFilter_Filter NO match with WQL || CQL
// MI_Filter Evaluate with WQL || CQL
// TODO: Split of split to reduce duplication?
NitsTest1(Test_IndicationFilter_Filter_NO_match, Test_IndicationFilter_SetupSubscribeReq_BothWays, NitsEmptyValue)
{
    MI_Boolean isMatch = MI_FALSE;
    CIM_InstDeletion indication;

    IndicationFilter* filter = IndicationFilter_New( subscribeReq );

    NitsAssertOrReturn(NULL != filter, PAL_T("Failed to initialize IndicationFilter with InvalidQueryLanguage"));

    if ( NitsAssert( MI_RESULT_OK == Instance_Construct( &indication.__instance, &CIM_InstDeletion_rtti, NULL ), PAL_T("Instance_Construct failed") ) )
    {
        MI_Value value;
        value.string = (MI_Char*)PAL_T("MSFT:DeletionIndicationIdentifier");
        if ( ! NitsAssert(MI_RESULT_OK == MI_Instance_SetElement( &indication.__instance, PAL_T("IndicationIdentifier"), &value, MI_STRING, 0 ), PAL_T("Set key failed on indication")) )
        {
            IndicationFilter_Destroy( filter );
            NitsReturn;
        }

        // Test the instance against the internal version of the call
        NitsAssert( MI_RESULT_OK == IndicationFilter_Filter( filter, &indication.__instance, &isMatch ), PAL_T("Filter failed") );
        NitsAssert( MI_FALSE == isMatch, PAL_T("Filter match when non-match expected") );

        // Also test same instance via external MI call
        NitsAssert( MI_RESULT_OK == MI_Filter_Evaluate( &filter->base, &indication.__instance, &isMatch ), PAL_T("Filter failed") );
        NitsAssert( MI_FALSE == isMatch, PAL_T("Filter match when non-match expected") ); NitsIgnoringError(); //negative test case and same error is returned when OOM or not found error occurs in WQL_LookupInstanceProperty

        NitsAssert( MI_RESULT_OK == MI_Instance_Destruct( &indication.__instance ), PAL_T("indication destruction failed") );
    }
    
    
    IndicationFilter_Destroy( filter );
    NitsAssert(0xFFFFFFFF == filter->magic, PAL_T("IndicationFilter: magic # not reset following destroy"));
    
}
NitsEndTest


// MI_Filter GetExpression
NitsTest1(Test_IndicationFilter_GetExpression, Test_IndicationFilter_SetupSubscribeReq, NitsEmptyValue)
{
    IndicationFilter* filter = IndicationFilter_New( subscribeReq );

    NitsAssertOrReturn(NULL != filter, PAL_T("Failed to initialize IndicationFilter with InvalidQueryLanguage"));

    MI_Filter* miFilter = &filter->base;
    const MI_Char* queryLang = NULL;
    const MI_Char* queryExpr = NULL;
    NitsAssert( MI_RESULT_OK == MI_Filter_GetExpression( miFilter, &queryLang, &queryExpr ), PAL_T("MI_Filter_GetExpression failed") );

    NitsAssert( 0 == Tcsncmp( queryLang, indicationLanguageWql, Tcslen(indicationLanguageWql) ), PAL_T("Unmatched query language") );
    NitsAssert( 0 == Tcsncmp( queryExpr, defaultIndicationFilter, Tcslen(defaultIndicationFilter) ), PAL_T("Unmatched query string") );

    IndicationFilter_Destroy( filter );

    NitsAssert(0xFFFFFFFF == filter->magic, PAL_T("IndicationFilter: magic # not reset following destroy"));
}
NitsEndTest

